/*
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#define DEBUG 0

#if DEBUG
#define DEBUG_PRINTF(...)  usart_write(__VA_ARGS__)
#else
#define DEBUG_PRINTF(...)
#endif
#define PRINTF(...)  usart_write(__VA_ARGS__)


#include "ubasic.h"
#include "tokenizer.h"
#include "ubasic_config.h"

#include "usart.h"
#include <avr/eeprom.h>
#include <util/delay.h>


#include <stdlib.h> /* exit() */

static char const *program_ptr;
#define MAX_STRINGLEN 40
static char string[MAX_STRINGLEN];

#define MAX_GOSUB_STACK_DEPTH 10
static int gosub_stack[MAX_GOSUB_STACK_DEPTH];
static int gosub_stack_ptr;

struct for_state {
  int line_after_for;
  int for_variable;
  int to;
  int step;
  char downto;
};
#define MAX_FOR_STACK_DEPTH 4
static struct for_state for_stack[MAX_FOR_STACK_DEPTH];
static int for_stack_ptr;

#define MAX_VARNUM 26
int variables[MAX_VARNUM];

static int ended;

static int current_linenum;

static int expr(void);
static void line_statement(void);
static void statement(void);
/*---------------------------------------------------------------------------*/
void
ubasic_init(const char *program)
{
  program_ptr = program;
  for_stack_ptr = gosub_stack_ptr = 0;
  tokenizer_init(program);
  ended = 0;
}
/*---------------------------------------------------------------------------*/
void
ubasic_break(void)
{
	#if BREAK_NOT_EXIT
	// zum Ende des Basic-Quelltextes gehen
	tokenizer_goto_end();
	#else
	exit(1);
	#endif
}

/*---------------------------------------------------------------------------*/
static void
accept(int token)
{
  if(token != tokenizer_token()) {
    DEBUG_PRINTF("Token not what was expected (expected %i, got %i)\r\n",
		 token, tokenizer_token());
    tokenizer_error_print(current_linenum);
    ubasic_break();
  }
  DEBUG_PRINTF("Expected %i, got it\r\n", token);
  tokenizer_next();
}
/*---------------------------------------------------------------------------*/
static int
varfactor(void)
{
  int r;
  DEBUG_PRINTF("varfactor: obtaining %i from variable %i\r\n", variables[tokenizer_variable_num()], tokenizer_variable_num());
  r = ubasic_get_variable(tokenizer_variable_num());
  accept(TOKENIZER_VARIABLE);
  return r;
}
/*---------------------------------------------------------------------------*/
static int
factor(void)
{
  int r=0;
  #if AVR_IN || AVR_ADC || AVR_EPEEK 
  int adr;
  char port, pin;
  #endif
  #if AVR_RND
  int b;
  #endif
  
  DEBUG_PRINTF("factor: token %i\n\r", tokenizer_token());
  switch(tokenizer_token()) {
  case TOKENIZER_NUMBER:
    r = tokenizer_num();
    accept(TOKENIZER_NUMBER);
    break;
    
  //?????
  case TOKENIZER_MINUS:
    accept(TOKENIZER_MINUS);
    r = tokenizer_num()*(-1);
    accept(TOKENIZER_NUMBER);
    break;

  case TOKENIZER_LEFTPAREN:
    accept(TOKENIZER_LEFTPAREN);
    r = expr();
    accept(TOKENIZER_RIGHTPAREN);
    break;
  
  #if AVR_RND
  case TOKENIZER_RND:
    accept(TOKENIZER_RND);
    accept(TOKENIZER_LEFTPAREN);
    b = expr();
    //while((r = rand()) >= RAND_MAX - (RAND_MAX % b)); 
	//r = r % b;
	r = rand();
    accept(TOKENIZER_RIGHTPAREN);
    break;
  #endif
  
  #if UBASIC_ABS
  case TOKENIZER_ABS:
    accept(TOKENIZER_ABS);
    accept(TOKENIZER_LEFTPAREN);
    r = expr();
    if (r<0) r=r*(-1);
    accept(TOKENIZER_RIGHTPAREN);
    break;
  #endif

  #if UBASIC_NOT
  case TOKENIZER_NOT:
    accept(TOKENIZER_NOT);
    accept(TOKENIZER_LEFTPAREN);
    r = ~expr();
    accept(TOKENIZER_RIGHTPAREN);
    break;
  #endif
    
  #if AVR_EPEEK
  case TOKENIZER_EPEEK:
    accept(TOKENIZER_EPEEK);
    accept(TOKENIZER_LEFTPAREN);
    adr = expr();
    r = eeprom_read_byte((unsigned char *)adr);
    accept(TOKENIZER_RIGHTPAREN);
    break;
  #endif
  
  #if AVR_ADC
  case TOKENIZER_ADC:
    accept(TOKENIZER_ADC);
    accept(TOKENIZER_LEFTPAREN);
	pin=tokenizer_num();
    tokenizer_next();
	if ((pin<0)||(pin>ADC_COUNT_MAX)) {
		//Fehlerfall
		DEBUG_PRINTF("adc_token: unknown channel %c\r\n", pin);
	    tokenizer_error_print(current_linenum);
		ubasic_break();
	} else {
		// ADC_Kanal und Referenzspannung (hier Avcc) einstellen
		ADMUX =  pin;
		ADMUX |= (1<<REFS0);
		// Prescaler 8:1 und ADC aktivieren
		ADCSRA = (1<<ADEN) | (1<<ADPS1) | (1<<ADPS0);
		// eine ADC-Wandlung fuer eine Dummy-Messung
		ADCSRA |= (1<<ADSC);              
		while (ADCSRA & (1<<ADSC));
		r = ADCW;
		// eigentliche Messung
		ADCSRA |= (1<<ADSC);  
		while (ADCSRA & (1<<ADSC));
		r = ADCW;
		// ADC wieder ausschalten
		ADCSRA=0;
	}	
    accept(TOKENIZER_RIGHTPAREN);
    break;
  #endif
    
  #if AVR_IN
  case TOKENIZER_IN:
    accept(TOKENIZER_IN);
    accept(TOKENIZER_LEFTPAREN);
	port=tokenizer_letter();
	tokenizer_next();
	accept(TOKENIZER_COMMA);
	pin=tokenizer_num();
	tokenizer_next();
    accept(TOKENIZER_RIGHTPAREN);
		switch (port) {
		#if HAVE_PORTA
		case 'a':
			if (bit_is_clear(PINA, pin)) r=0; else r=1;
			break;
		#endif
		#if HAVE_PORTB
		case 'b':
			if (bit_is_clear(PINB, pin)) r=0; else r=1;
			break;
		#endif
		#if HAVE_PORTC
		case 'c':
			if (bit_is_clear(PINC, pin)) r=0; else r=1;
			break;
		#endif
		#if HAVE_PORTD
		case 'd':
			if (bit_is_clear(PIND, pin)) r=0; else r=1;
			break;
		#endif
		default:
			DEBUG_PRINTF("in_token: unknown port %c\r\n", port);
		    tokenizer_error_print(current_linenum);
			ubasic_break();
		}
    break;
  #endif
  
  default:
    r = varfactor();
    break;
  }
  return r;
}
/*---------------------------------------------------------------------------*/
static int
term(void)
{
  int f1, f2;
  int op;

  f1 = factor();
  op = tokenizer_token();
  DEBUG_PRINTF("term: token %i\n\r", op);
  while(op == TOKENIZER_ASTR ||
	op == TOKENIZER_SLASH    ||
	op == TOKENIZER_MOD) {
    tokenizer_next();
    f2 = factor();
    DEBUG_PRINTF("term: %i %i %i\r\n", f1, op, f2);
    switch(op) {
    case TOKENIZER_ASTR:
      f1 = f1 * f2;
      break;
    case TOKENIZER_SLASH:
      f1 = f1 / f2;
      break;
    case TOKENIZER_MOD:
      f1 = f1 % f2;
      break;
    }
    op = tokenizer_token();
  }
  DEBUG_PRINTF("term: %i\r\n", f1);
  return f1;
}
/*---------------------------------------------------------------------------*/
static int
expr(void)
{
  int t1, t2;
  int op;
  
  t1 = term();
  op = tokenizer_token();
  DEBUG_PRINTF("expr: token %i\r\n", op);
  while(op == TOKENIZER_PLUS ||
	op == TOKENIZER_MINUS ||
	op == TOKENIZER_AND ||
	op == TOKENIZER_OR) {
    tokenizer_next();
    t2 = term();
    DEBUG_PRINTF("expr: %i %i %i\r\n", t1, op, t2);
    switch(op) {
    case TOKENIZER_PLUS:
      t1 = t1 + t2;
      break;
    case TOKENIZER_MINUS:
      t1 = t1 - t2;
      break;
    case TOKENIZER_AND:
      t1 = t1 & t2;
      break;
    case TOKENIZER_OR:
      t1 = t1 | t2;
      break;
    }
    op = tokenizer_token();
  }
  DEBUG_PRINTF("expr: %i\r\n", t1);
  return t1;
}
/*---------------------------------------------------------------------------*/
static int
relation(void)
{
  int r1, r2;
  int op;
  
  r1 = expr();
  op = tokenizer_token();
  DEBUG_PRINTF("relation: token %i\r\n", op);
  while(op == TOKENIZER_LT ||
	op == TOKENIZER_GT ||
	op == TOKENIZER_EQ) {
    tokenizer_next();
    r2 = expr();
    DEBUG_PRINTF("relation: %i %i %i\r\n", r1, op, r2);
    switch(op) {
    case TOKENIZER_LT:
      r1 = r1 < r2;
      break;
    case TOKENIZER_GT:
      r1 = r1 > r2;
      break;
    case TOKENIZER_EQ:
      r1 = r1 == r2;
      break;
    }
    op = tokenizer_token();
  }
  return r1;
}
/*---------------------------------------------------------------------------*/
static void
jump_linenum(int linenum)
{
  tokenizer_init(program_ptr);
  while(tokenizer_num() != linenum) {
    do {
      do {
	tokenizer_next();
      } while(tokenizer_token() != TOKENIZER_CR &&
	      tokenizer_token() != TOKENIZER_ENDOFINPUT);
      if(tokenizer_token() == TOKENIZER_CR) {
	tokenizer_next();
      }
    } while(tokenizer_token() != TOKENIZER_NUMBER);
    DEBUG_PRINTF("jump_linenum: Found line %i\r\n", tokenizer_num());
  }
}
/*---------------------------------------------------------------------------*/
static void
goto_statement(void)
{
  accept(TOKENIZER_GOTO);
  jump_linenum(tokenizer_num());
}
/*---------------------------------------------------------------------------*/
static void
print_statement(void)
{
  accept(TOKENIZER_PRINT);
  do {
    DEBUG_PRINTF("Print loop\r\n");
    if(tokenizer_token() == TOKENIZER_STRING) {
      tokenizer_string(string, sizeof(string));
      PRINTF("%s", string);
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_COMMA) {
      PRINTF(" ");
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_SEMICOLON) {
      tokenizer_next();
    } else if(tokenizer_token() == TOKENIZER_VARIABLE ||
	          #if AVR_RND
			  tokenizer_token() == TOKENIZER_RND      ||
			  #endif
	          #if UBASIC_ABS
			  tokenizer_token() == TOKENIZER_ABS      ||
			  #endif
	          #if UBASIC_NOT
			  tokenizer_token() == TOKENIZER_NOT      ||
			  #endif
	          #if AVR_EPEEK
			  tokenizer_token() == TOKENIZER_EPEEK    ||
			  #endif
			  #if AVR_IN
			  tokenizer_token() == TOKENIZER_IN       ||
			  #endif
			  #if AVR_ADC
			  tokenizer_token() == TOKENIZER_ADC      ||
			  #endif
	          tokenizer_token() == TOKENIZER_NUMBER   
			  ) {
      PRINTF("%i", expr());
    } else {
      break;
    }
  } while(tokenizer_token() != TOKENIZER_CR &&
	  tokenizer_token() != TOKENIZER_ENDOFINPUT);
  PRINTF("\n\r");
  DEBUG_PRINTF("End of print\r\n");
  tokenizer_next();
}
/*---------------------------------------------------------------------------*/
static void
if_statement(void)
{
  int r;
  
  accept(TOKENIZER_IF);

  r = relation();
  DEBUG_PRINTF("if_statement: relation %i\r\n", r);
  accept(TOKENIZER_THEN);
  if(r) {
    statement();
    do {
      tokenizer_next();
    } while(tokenizer_token() != TOKENIZER_CR &&
	        tokenizer_token() != TOKENIZER_ENDOFINPUT);
	if(tokenizer_token() == TOKENIZER_CR) tokenizer_next();
  } else {
    do {
      tokenizer_next();
    } while(tokenizer_token() != TOKENIZER_ELSE &&
	        tokenizer_token() != TOKENIZER_CR &&
	        tokenizer_token() != TOKENIZER_ENDOFINPUT);
    if(tokenizer_token() == TOKENIZER_ELSE) {
      tokenizer_next();
      statement();
    } else if(tokenizer_token() == TOKENIZER_CR) {
      tokenizer_next();
    }
  }
}
/*---------------------------------------------------------------------------*/
static void
let_statement(void)
{
  int var;

  var = tokenizer_variable_num();

  accept(TOKENIZER_VARIABLE);
  accept(TOKENIZER_EQ);
  ubasic_set_variable(var, expr());
  DEBUG_PRINTF("let_statement: assign %i to %i\r\n", variables[var], var);
  tokenizer_next();
}
/*---------------------------------------------------------------------------*/
static void
gosub_statement(void)
{
  int linenum;
  accept(TOKENIZER_GOSUB);
  linenum = tokenizer_num();
  accept(TOKENIZER_NUMBER);
  tokenizer_next();
  if(gosub_stack_ptr < MAX_GOSUB_STACK_DEPTH) {
    gosub_stack[gosub_stack_ptr] = tokenizer_num();
    gosub_stack_ptr++;
    jump_linenum(linenum);
  } else {
    DEBUG_PRINTF("gosub_statement: gosub stack exhausted\r\n");
  }
}
/*---------------------------------------------------------------------------*/
static void
return_statement(void)
{
  accept(TOKENIZER_RETURN);
  if(gosub_stack_ptr > 0) {
    gosub_stack_ptr--;
    jump_linenum(gosub_stack[gosub_stack_ptr]);
  } else {
    DEBUG_PRINTF("return_statement: non-matching return\r\n");
  }
}
/*---------------------------------------------------------------------------*/
static void
next_statement(void)
{
  int var;
  
  accept(TOKENIZER_NEXT);
  var = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  if(for_stack_ptr > 0 && var == for_stack[for_stack_ptr - 1].for_variable) {
    ubasic_set_variable(var, ubasic_get_variable(var) + for_stack[for_stack_ptr - 1].step);
    if(((ubasic_get_variable(var) <= for_stack[for_stack_ptr - 1].to) && !for_stack[for_stack_ptr - 1].downto)||
       ((ubasic_get_variable(var) >= for_stack[for_stack_ptr - 1].to) && for_stack[for_stack_ptr - 1].downto)
      ) {
      jump_linenum(for_stack[for_stack_ptr - 1].line_after_for);
    } else {
      for_stack_ptr--;
      accept(TOKENIZER_CR);
    }
  } else {
    DEBUG_PRINTF("next_statement: non-matching next (expected %i, found %i)\r\n", for_stack[for_stack_ptr - 1].for_variable, var);
    accept(TOKENIZER_CR);
  }

}
/*---------------------------------------------------------------------------*/
static void
for_statement(void)
{
  int for_variable, to, step; 
  char downto;
  
  accept(TOKENIZER_FOR);
  for_variable = tokenizer_variable_num();
  accept(TOKENIZER_VARIABLE);
  accept(TOKENIZER_EQ);
  ubasic_set_variable(for_variable, expr());
  if (tokenizer_token() == TOKENIZER_TO) {
  	downto = 0;
  } else if (tokenizer_token() == TOKENIZER_DOWNTO) {
  	downto = 1;
  } else tokenizer_error_print(current_linenum);
  tokenizer_next();
  to = expr();
  if(tokenizer_token() == TOKENIZER_STEP) {
  	tokenizer_next();
  	step = expr();
  } else {
  	step = 1;
  }
  if (downto) step *= -1;
  accept(TOKENIZER_CR);

  if(for_stack_ptr < MAX_FOR_STACK_DEPTH) {
    for_stack[for_stack_ptr].line_after_for = tokenizer_num();
    for_stack[for_stack_ptr].for_variable = for_variable;
    for_stack[for_stack_ptr].to = to;
    for_stack[for_stack_ptr].step = step;
    for_stack[for_stack_ptr].downto = downto;
    
    DEBUG_PRINTF("for_statement: new for, var %i to %i\r\n",
		 for_stack[for_stack_ptr].for_variable,
		 for_stack[for_stack_ptr].to);
		 
    for_stack_ptr++;
  } else {
    DEBUG_PRINTF("for_statement: for stack depth exceeded\r\n");
  }
}
/*---------------------------------------------------------------------------*/
static void
end_statement(void)
{
  accept(TOKENIZER_END);
  ended = 1;
}

/*---------------------------------------------------------------------------*/
#if AVR_EPOKE
static void
epoke_statement(void)
{
	int adr, val;
	accept(TOKENIZER_EPOKE);
    accept(TOKENIZER_LEFTPAREN);
    adr = expr();
    accept(TOKENIZER_RIGHTPAREN);
    accept(TOKENIZER_EQ);
	val = expr();
	eeprom_write_byte((unsigned char *)adr, val);
	tokenizer_next();
}
#endif

/*---------------------------------------------------------------------------*/
// dir(<port>, <pin>)=<val|var|expr>
#if AVR_DIR
static void
dir_statement(void)
{
	char port, pin, val;
	accept(TOKENIZER_DIR);
    accept(TOKENIZER_LEFTPAREN);
	port=tokenizer_letter();
	tokenizer_next();
	accept(TOKENIZER_COMMA);
	pin=tokenizer_num();
	tokenizer_next();
    accept(TOKENIZER_RIGHTPAREN);
    accept(TOKENIZER_EQ);
	if (expr()) val=1; else val=0;
	switch (port) {
		#if HAVE_PORTA
		case 'a':
			if (val) DDRA |= (1 << pin); else DDRA &= ~(1 << pin);
			break;
		#endif
		#if HAVE_PORTB
		case 'b':
			if (val) DDRB |= (1 << pin); else DDRB &= ~(1 << pin);
			break;
		#endif
		#if HAVE_PORTC
		case 'c':
			if (val) DDRC |= (1 << pin); else DDRC &= ~(1 << pin);
			break;
		#endif
		#if HAVE_PORTD
		case 'd':
			if (val) DDRD |= (1 << pin); else DDRD &= ~(1 << pin);
			break;
		#endif
		default:
			DEBUG_PRINTF("dir_statement: unknown port %c\r\n", port);
		    tokenizer_error_print(current_linenum);
			ubasic_break();
	}
	tokenizer_next();
}
#endif

/*---------------------------------------------------------------------------*/
// out(<port>, <pin>)=<val|var|expr>
#if AVR_OUT
static void
out_statement(void)
{
	char port, pin, val;
	accept(TOKENIZER_OUT);
    accept(TOKENIZER_LEFTPAREN);
	port=tokenizer_letter();
	tokenizer_next();
	accept(TOKENIZER_COMMA);
	pin=tokenizer_num();
	tokenizer_next();
    accept(TOKENIZER_RIGHTPAREN);
    accept(TOKENIZER_EQ);
	val = expr();
	switch (port) {
		#if HAVE_PORTA
		case 'a':
			if (val) PORTA |= (1 << pin); else PORTA &= ~(1 << pin);
			break;
		#endif
		#if HAVE_PORTB
		case 'b':
			if (val) PORTB |= (1 << pin); else PORTB &= ~(1 << pin);
			break;
		#endif
		#if HAVE_PORTC
		case 'c':
			if (val) PORTC |= (1 << pin); else PORTC &= ~(1 << pin);
			break;
		#endif
		#if HAVE_PORTD
		case 'd':
			if (val) PORTD |= (1 << pin); else PORTD &= ~(1 << pin);
			break;
		#endif
		default:
			DEBUG_PRINTF("out_statement: unknown port %c\r\n", port);
			tokenizer_error_print(current_linenum);
			ubasic_break();
	}
	tokenizer_next();
}
#endif

/*---------------------------------------------------------------------------*/
#if AVR_WAIT
static void
wait_statement(void)
{
	int delay;
	accept(TOKENIZER_WAIT);
	delay=expr();
	for (int i=0; i<delay; i++) _delay_ms(1);
	tokenizer_next();

}
#endif

/*---------------------------------------------------------------------------*/
#if AVR_RND
static void
srand_statement(void)
{
	uint16_t seed = 0;
	uint16_t *p = (uint16_t*) (RAMEND+1);
	extern uint16_t __heap_start;
	accept(TOKENIZER_SRND);
	while (p >= &__heap_start + 1)
		seed ^= * (--p);
	srand(seed);
	tokenizer_next();
}
#endif

/*---------------------------------------------------------------------------*/
static void
statement(void)
{
  int token;
  
  token = tokenizer_token();
  
  switch(token) {
  case TOKENIZER_PRINT:
    print_statement();
    break;
  case TOKENIZER_IF:
    if_statement();
    break;
  case TOKENIZER_GOTO:
    goto_statement();
    break;
  case TOKENIZER_GOSUB:
    gosub_statement();
    break;
  case TOKENIZER_RETURN:
    return_statement();
    break;
  case TOKENIZER_FOR:
    for_statement();
    break;
  case TOKENIZER_NEXT:
    next_statement();
    break;
  case TOKENIZER_END:
    end_statement();
    break;
  case TOKENIZER_LET:
    accept(TOKENIZER_LET);
    /* Fall through. */
  case TOKENIZER_VARIABLE:
    let_statement();
    break;
	
  #if AVR_EPOKE
  case TOKENIZER_EPOKE:
    epoke_statement();
    break;
  #endif

  #if AVR_WAIT
  case TOKENIZER_WAIT:
    wait_statement();
    break;
  #endif
  
  #if AVR_DIR
  case TOKENIZER_DIR:
    dir_statement();
    break;
  #endif

  #if AVR_OUT
  case TOKENIZER_OUT:
    out_statement();
    break;
  #endif
  
  #if AVR_RND
  case TOKENIZER_SRND:
    srand_statement();
    break;
  #endif

  default:
    PRINTF("default von statement()\r\n");
    DEBUG_PRINTF("ubasic.c: statement(): not implemented %i\r\n", token);
    tokenizer_error_print(current_linenum);
    ubasic_break();
  }
}
/*---------------------------------------------------------------------------*/
static void
line_statement(void)
{
  DEBUG_PRINTF("----------- Line number %i ---------\r\n", tokenizer_num());
  current_linenum = tokenizer_num();
  accept(TOKENIZER_NUMBER);
  statement();
  return;
}
/*---------------------------------------------------------------------------*/
void
ubasic_run(void)
{
  if(tokenizer_finished()) {
    DEBUG_PRINTF("uBASIC program finished\r\n");
    return;
  }

  line_statement();
}
/*---------------------------------------------------------------------------*/
int
ubasic_finished(void)
{
  return ended || tokenizer_finished();
}
/*---------------------------------------------------------------------------*/
void
ubasic_set_variable(int varnum, int value)
{
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    variables[varnum] = value;
  }
}
/*---------------------------------------------------------------------------*/
int
ubasic_get_variable(int varnum)
{
  if(varnum >= 0 && varnum <= MAX_VARNUM) {
    return variables[varnum];
  }
  return 0;
}
/*---------------------------------------------------------------------------*/
